/*
 * MentaAffinity => http://mentaaffinity.soliveirajr.com
 * Copyright (C) 2012  Sergio Oliveira Jr. (sergio.oliveira.jr@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package org.mentaaffinity;

import static org.mentalog.Log.*;

import java.io.File;
import java.util.List;

/**
 * 
 * The main class to set and control many different affinity strategies.
 * 
 * @author Sergio Oliveira Jr.
 */
public class Affinity {

	private static boolean enabled = true;
	private static Boolean available = null;
	
	static {
		if (isAvailable()) {
			init();
			Debug.log("MentaAffinity loaded!", CpuInfo.instance().toString());
		}
	}
	
	private static Chip getSingleChip() {
		List<Chip> chips = CpuInfo.getChips();
		if (chips.size() != 1) {
			throw new AffinityException("Your computer has more than one chip so you must specify the chip you want to use when assigning threads!");
		}
		return chips.get(0);
	}

	private static Chip getChip(int id) {
		for (int i = 0; i < CpuInfo.getChips().size(); i++) {
			Chip c = CpuInfo.getChips().get(i);
			if (c.getId() == id) {
				return c;
			}
		}
		throw new AffinityException("Cannot find chip for id: " + id);
	}

	/**
	 * By default affinity is enabled but you can turn it off if you want to, 
	 * for example if you are running in a platform other than Linux. The effect
	 * of turning it off is that no affinity will be set for any thread, in other words,
	 * bind and unbind will be ignored.
	 * 
	 * @param flag true to enable
	 */
	public static void setEnable(boolean flag) {
		Affinity.enabled = flag;
	}

	private synchronized static Processor getProcessor(int id) {
		List<Chip> chips = CpuInfo.getChips();
		for (int i = 0; i < chips.size(); i++) {
			List<Core> cores = chips.get(i).getCores();
			for (int j = 0; j < cores.size(); j++) {
				List<Processor> procs = cores.get(j).getProcessors();
				for (int k = 0; k < procs.size(); k++) {
					if (procs.get(k).getId() == id) {
						return procs.get(k);
					}
				}
			}
		}
		return null;
	}

	/**
	 * Checks if MentaAffinity is available for your JVM. MentaAffinity can only
	 * work if JNA is available and if a /proc/cpuinfo exists.
	 * 
	 * @return true if MentaAffinity is available
	 */
	public final static boolean isAvailable() {
		if (available == null) {
			try {
				Class.forName("com.sun.jna.Platform");
			} catch (Exception ignored) {
				Debug.log("MentaAffinity", "Cannot load JNA! Make sure to install from https://github.com/twall/jna. You may have to include the jna.jar in the classpath.");
				available = false;
				return false;
			}
			
			File file = new File(CpuInfo.DEFAULT_CPUINFO);
			if (!file.exists()) {
				Debug.log("MentaAffinity", "Cannot find file:", CpuInfo.DEFAULT_CPUINFO);
				available = false;
				return false;
			}
			available = true;
		}
		return available;
	}

	/**
	 * Initialize MentaAffinity by loading the /proc/cpuinfo file.
	 */
	public static void init() {
		CpuInfo.init();
		
	}

	/**
	 * Assign the thread to a single logical processor.
	 * 
	 * @param procId the id of the processor (starts at 0).
	 * @param t the thread
	 * @throws AffinityException if cannot assign to this processor for whatever reason
	 */
	public synchronized static void assignToProcessor(int procId, Thread t) {
		Processor p = getProcessor(procId);
		if (p == null) {
			throw new AffinityException("Cannot find processor: " + procId);
		}
		p.assign(t);
	}

	/**
	 * Choose the first free logical processor available and assign the thread to this processor.
	 * 
	 * @param t the thread
	 * @return the processor to where the thread was bound
	 * @throws AffinityException if cannot assign to this processor for whatever reason
	 */
	public synchronized static Processor assignToAnyFreeProcessor(Thread t) {
		return assignToAnyFreeProcessor(getSingleChip().getId(), t);
	}

	/**
	 * Choose the first free logical processor available in that chip and assign the thread to it.
	 * 
	 * @param chipId the physical processor we want to assign to
	 * @param t the thread we want to assign
	 * @return the processor that this thread was assigned to
	 * @throws AffinityException if cannot assign to this processor for whatever reason
	 */
	public synchronized static Processor assignToAnyFreeProcessor(int chipId, Thread t) {
		Chip chip = getChip(chipId);
		Core core = chip.getFreeCore();
		if (core == null) {
			throw new AffinityException("Cannot find a free core from chip: " + chip);
		}
		Processor processor = core.getFreeProcessor();
		processor.assign(t);
		return processor;
	}

	/**
	 * Looks for a core that has no threads running and assign this thread to it. This is very useful
	 * for chips with hyper-threading when you want to isolate a single thread in the WHOLE core so
	 * there is nobody else competing for the L1 cache in the other logical processor. MentaAffinity
	 * will mark the other logical processor as idle until the thread unbinds from the core.
	 * 
	 * @param t the thread to assign
	 * @return the core to where this thread was assigned
	 * @throws AffinityException if cannot assign to this core for whatever reason
	 */
	public synchronized static Core assignToAnyFreeCore(Thread t) {
		return assignToAnyFreeCore(getSingleChip().getId(), t);
	}

	/**
	 * Same thing as assignToAnyFreeCore(Thread t) except that you provide the chip
	 * from where to get a core. Useful for architectures with more than one physical processor.
	 * 
	 * @param chipId the physical processor
	 * @param t the thread to assign
	 * @return the core to where the thread was assigned
	 * @throws AffinityException if cannot assign to this core for whatever reason
	 */
	public synchronized static Core assignToAnyFreeCore(int chipId, Thread t) {
		Chip chip = getChip(chipId);
		Core core = chip.getFullyFreeCore();
		if (core == null) {
			throw new AffinityException("Cannot find a fully free core from chip: " + chip);
		}
		core.assign(t);
		return core;
	}

	/**
	 * Finds a free core and assigns both threads to both of its logical processors.
	 * Of course if the chip does not support hyper-threading this will throw an exception.
	 * 
	 * @param t1 the first thread to assign
	 * @param t2 the seconds thread to assign
	 * @throws AffinityException if cannot assign to this core for whatever reason
	 */
	public synchronized static void assignToSameCore(Thread t1, Thread t2) {
		assignToSameCore(getSingleChip().getId(), t1, t2);
	}

	/**
	 * Same thing as assignToSameCore(Thread t1, Thread t2) except that you can provide
	 * the physical processor. Useful for machines with more than one physical processor.
	 * 
	 * @param chipId the physical processor
	 * @param t1 the first thread to assign
	 * @param t2 the second thread to assign
	 * @throws AffinityException if cannot assign to this core for whatever reason
	 */
	public synchronized static void assignToSameCore(int chipId, Thread t1, Thread t2) {
		Chip chip = getChip(chipId);
		Core core = chip.getFullyFreeCore();
		if (core == null) {
			throw new AffinityException("Cannot find a fully free core from chip: " + chip);
		}
		core.assign(t1, t2);
	}

	/**
	 * Make sure both threads will NOT be assigned to the same core. Note that they will still
	 * be in the same chip. If you want them to be in separate chips use the assignToDifferentChips method.
	 * 
	 * @param t1 the first thread
	 * @param t2 the second thread
	 * @throws AffinityException if cannot assign the threads for whatever reason
	 */
	public synchronized static void assignToDifferentCores(Thread t1, Thread t2) {
		assignToDifferentCores(getSingleChip().getId(), t1, t2);
	}

	/**
	 * Same thing as assignToDifferentCores(Thread t1, Thread t2) except that you can provide
	 * the physical processor. Useful if the machine has more than one physical processor.
	 * 
	 * @param chipId the physical processor
	 * @param t1 the first thread
	 * @param t2 the second thread
	 * @throws AffinityException if cannot assign the threads for whatever reason
	 */
	public synchronized static void assignToDifferentCores(int chipId, Thread t1, Thread t2) {
		Chip chip = getChip(chipId);
		if (chip.getNumberOfFreeCores() >= 2) {
			Core core1 = chip.getFreeCore();
			core1.assign(t1);
			Core core2 = chip.getFreeCore();
			core2.assign(t2);
		} else {
			throw new AffinityException("There are no two free cores in this chip: " + chip);
		}
	}

	/**
	 * Make sure that both threads are assigned to separate physical processors.
	 * 
	 * @param t1 the first thread
	 * @param t2 the second thread
	 * @throws AffinityException if cannot assign the threads for whatever reason
	 */
	public synchronized static void assignToDifferentChips(Thread t1, Thread t2) {
		List<Chip> chips = CpuInfo.getChips();
		if (chips.size() == 1) {
			throw new AffinityException("Your computer only have one chip!");
		}
		if (chips.size() > 2) {
			throw new AffinityException("Your computer has more than one chip so you must specify the chip you want to use when assigning threads!");
		}
		assignToDifferentChips(chips.get(0).getId(), t1, chips.get(1).getId(), t2);
	}

	/**
	 * Assign the threads to the provided physical processors. Useful if you have a machine with
	 * more than 2 physical processors.
	 * 
	 * @param chipId1 the chip to assign the first thread
	 * @param t1 the first thread
	 * @param chipId2 the chip to assign the second thread
	 * @param t2 the second thread
	 * @throws AffinityException if cannot assign the threads for whatever reason
	 */
	public synchronized static void assignToDifferentChips(int chipId1, Thread t1, int chipId2, Thread t2) {
		Chip chip1 = getChip(chipId1);
		Chip chip2 = getChip(chipId2);
		if (!chip1.hasFreeCore()) {
			throw new AffinityException("This chip does not have a free core: " + chip1);
		}
		if (!chip2.hasFreeCore()) {
			throw new AffinityException("This chip does not have a free core: " + chip2);
		}
		chip1.getFreeCore().getFreeProcessor().assign(t1);
		chip2.getFreeCore().getFreeProcessor().assign(t2);
	}

	/**
	 * This method actually makes the system call to set the thread affinity. <b>You must call this method from inside your thread, in other words, 
	 * from inside its run method.</b> If MentaAffinity is not available (no-JNA) or disabled then this method is ignored. Also if the thread was not
	 * assigned to anything then this method is ignored.
	 * 
	 * @throws AffinityException if it wasn't possible to bind by any reason
	 */
	public final static void bind() {
		
		if (!enabled || !isAvailable()) {
			return;
		}
		
		synchronized(Affinity.class) { 
    		Thread t = Thread.currentThread();
    		// first check assigned core...
    		Core core = Core.getAssignedCoreForThread(t);
    		if (core != null) {
    			// check which processor you have to bind...
    			for (int i = 0; i < core.getProcessors().size(); i++) {
    				Processor p = core.getProcessors().get(i);
    				if (p.isAssigned()) {
    					if (p.getThread() != t) {
    						throw new AffinityException("Cannot bind to this processor: " + p.getThread().getName());
    					}
    					try {
    						ThreadAffinity.setCurrentThreadAffinityMask(p.getMask());
    						if (Debug.isEnabled()) {
    							Debug.log("MentaAffinity", "Thread bound to core:", "thread=", p.getThread().getName(), "core=", core, "proc=", p);
    						}
    						return;
    					} catch (Exception e) {
    						throw new AffinityException("Cannot bind with native code!", e);
    					}
    				}
    			}
    
    			throw new AffinityException("Error binding to core: " + core);
    		}
    		// now check for processor
    		Processor processor = Processor.getAssignedProcessorForThread(t);
    		if (processor != null) {
    			if (processor.isAssigned()) {
    				if (processor.getThread() != t) {
    					throw new AffinityException("Cannot bind to this processor: " + processor.getThread().getName());
    				}
    				try {
    					ThreadAffinity.setCurrentThreadAffinityMask(processor.getMask());
    					if (Debug.isEnabled()) {
							Debug.log("MentaAffinity", "Thread bound to processor:", "thread=", processor.getThread().getName(), "proc=", processor);
						}
    					return;
    				} catch (Exception e) {
    					throw new AffinityException("Cannot bind with native code!", e);
    				}
    			}
    			throw new AffinityException("Error binding to processor: " + processor);
    		}

		// do nothing because I am not assigned to anything!
		}
	}

	/**
	 * This method must be called when the thread is about to die. This does not make any system call
	 * because the thread is dying anyways and there is no need to change its affinity.
	 * However this method is important as it frees (unassigns) the logical processors for other threads.
	 * 
	 * NOTE: This method will actually release the processor/core, so other threads can be assigned to it.
	 */
	public final static void unbind() {
		
		if (!enabled || !isAvailable()) {
			return;
		}
		
		synchronized(Affinity.class) {

    		Thread t = Thread.currentThread();
    		// first check assigned core...
    		Core core = Core.getAssignedCoreForThread(t);
    		if (core != null) {
    			core.unassign(t);
    			return;
    		}
    		// now check for processor...
    		Processor processor = Processor.getAssignedProcessorForThread(t);
    		if (processor != null) {
    			processor.unassign(t);
    			return;
    		}

    		// do nothing because I was not assigned to anything!
		}
	}

	/**
	 * Prints to System.out a nice outlook of the logical processors situation. It will also
	 * show you the machine architecture with chips, cores and processors.
	 */
	public synchronized static void printSituation() {
		List<Chip> chips = CpuInfo.getChips();
		for (int i = 0; i < chips.size(); i++) {
			Chip chip = chips.get(i);
			chip.printSituation();
		}
	}
	
	public static void main(String[] args) {
		
		// java -DlogLevel=debug -DlogColors=true -cp /usr/share/java/jna.jar:../MentaLog/target/classes/:target/classes org.mentaaffinity.Affinity
		
		System.out.println("JPA available: " + isAvailable());
		System.out.println();
		System.out.println("Processors:");
		System.out.println();
		Affinity.printSituation();
	}
}