/*
 * MentaAffinity => http://mentaaffinity.soliveirajr.com
 * Copyright (C) 2012  Sergio Oliveira Jr. (sergio.oliveira.jr@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package org.mentaaffinity;

import java.util.ArrayList;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;

/**
 * An object representing a processor core, for multi-core processors.
 * 
 * @author Sergio Oliveira Jr.
 */
public class Core {

	private static final Map<Thread, Core> threadToCore = Collections.synchronizedMap(new IdentityHashMap<Thread, Core>());

	private final int chipId;
	private final int coreId;
	private final List<Processor> processors;
	private final String name;

	public Core(int chipId, int coreId, ArrayList<Processor> processors) {
		if (processors == null || processors.size() == 0) {
			throw new AffinityException("A core must have at least one processor!");
		} else if (processors.size() > 2) {
			throw new AffinityException("Can only handle a max of 2 processors per core: " + processors.size());
		}
		this.chipId = chipId;
		this.coreId = coreId;
		this.processors = Collections.unmodifiableList(processors);
		this.name = "Core-" + coreId + "." + chipId;
	}

	public int getCoreId() {
		return coreId;
	}

	public int getChipId() {
		return chipId;
	}

	public List<Processor> getProcessors() {
		return processors;
	}

	public int getNumberOfProcessors() {
		return processors.size();
	}

	static Core getAssignedCoreForThread(Thread t) {
		return threadToCore.get(t);
	}

	synchronized void assign(Thread t1, Thread t2) {
		if (getNumberOfProcessors() != 2) {
			throw new AffinityException("The chip does not support hyperthreading!");
		}
		if (!isFullyFree()) {
			throw new AffinityException("This core is not fully free!");
		}
		processors.get(0).assign(t1);
		processors.get(1).assign(t2);
	}

	public synchronized boolean isFullyFree() {
		for (int i = 0; i < processors.size(); i++) {
			Processor p = processors.get(i);
			if (!p.isFree()) {
				return false;
			}
		}
		return true;
	}

	public synchronized boolean hasFreeProcessor() {
		return getFreeProcessor() != null;
	}

	public synchronized Processor getFreeProcessor() {
		for (int i = 0; i < processors.size(); i++) {
			Processor p = processors.get(i);
			if (p.isFree()) {
				return p;
			}
		}
		return null;
	}

	synchronized void assign(Thread thread) {
		if (!isFullyFree()) {
			throw new AffinityException("Core is not fully free!");
		}
		if (threadToCore.containsKey(thread)) {
			throw new AffinityException("Thread is already assigned to a core!");
		}
		if (processors.size() == 1) {
			throw new AffinityException("Your machine does not support hyper-threading!");
		} else {
			processors.get(0).assign(thread);
			processors.get(1).setIdle(thread, true);
		}
		threadToCore.put(thread, this);
	}

	synchronized void unassign(Thread thread) {
		if (!threadToCore.containsKey(thread)) {
			throw new AffinityException("Thread is not assigned to any core!");
		}

		if (processors.size() == 1) {
			processors.get(0).unassign(thread);
		} else {
			processors.get(0).unassign(thread);
			processors.get(1).setIdle(thread, false);
		}
		threadToCore.remove(thread);
	}

	@Override
	public int hashCode() {
		return chipId * 31 + coreId;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Core) {
			Core c = (Core) obj;
			return c.coreId == this.coreId && c.chipId == this.chipId;
		}
		return false;
	}

	@Override
	public String toString() {
		return name;
	}
}