/*
 * MentaAffinity => http://mentaaffinity.soliveirajr.com
 * Copyright (C) 2012  Sergio Oliveira Jr. (sergio.oliveira.jr@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package org.mentaaffinity;

import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.Map;

/**
 * An object representing a logical processor, the smallest processing unit capable of running only one thread at a time. <b>Do not confuse this class with a physical processor.</b>
 * 
 * @author Sergio Oliveira Jr.
 */
public class Processor {

	private static final Map<Thread, Processor> threadToProcessor = Collections.synchronizedMap(new IdentityHashMap<Thread, Processor>());

	private final int procId;
	private final int chipId;
	private final int coreId;
	private final boolean isExcluded;
	private boolean isAssigned = false;
	private boolean leaveIdle = false;
	private Thread thread = null;
	private final String name;

	public Processor(final int procId, final int chipId, final int coreId, final boolean isExcluded) {
		this.procId = procId;
		this.chipId = chipId;
		this.coreId = coreId;
		this.isExcluded = isExcluded;
		this.name = "Processor-" + procId;
	}

	public synchronized String getStatusString() {
		if (isExcluded) {
			return "reserved for the OS";
		} else if (leaveIdle) {
			return "left idle by " + thread.getName();
		} else if (isAssigned) {
			if (thread.isAlive()) {
				return "bound to " + thread.getName() + " (running)";	
			} else {
				return "assigned to " + thread.getName() + " (not-started)";
			}
		} else {
			return "free";
		}
	}

	public int getId() {
		return procId;
	}

	public int getChipId() {
		return chipId;
	}

	public int getCoreId() {
		return coreId;
	}

	public boolean isExcluded() {
		return isExcluded;
	}

	public synchronized boolean isFree() {
		return !isAssigned && !isExcluded && !leaveIdle;
	}

	public synchronized boolean isAssigned() {
		return isAssigned;
	}

	public synchronized Thread getThread() {
		return thread;
	}

	synchronized void assign(Thread thread) {
		// check if re-assigning before starting:
		Processor previousProc = threadToProcessor.get(thread);
		if (previousProc != null && !thread.isAlive()) {
			// re-assigning before start => allow!
			// but first unassign the previous proc from the previous processor...
			previousProc.unassign(thread);
		} else {
			if (isAssigned) {
				throw new AffinityException("Processor is already assigned to a thread: " + this.thread);
			}
			if (threadToProcessor.containsKey(thread)) {
				throw new AffinityException("Thread is already assigned to a processor!");
			}
		}
		this.isAssigned = true;
		this.thread = thread;
		threadToProcessor.put(thread, this);
	}

	synchronized void setIdle(Thread thread, boolean flag) {
		if (!flag) {
			if (this.thread != thread) {
				throw new AffinityException("Processor was not made idle by this thread: " + this.thread);
			}
			this.leaveIdle = false;
			this.thread = null;
		} else {
			if (this.leaveIdle) {
				throw new AffinityException("Processor is already idle: " + this.thread);
			}
			this.leaveIdle = true;
			this.thread = thread;
		}
	}

	synchronized void unassign(Thread thread) {
		if (this.thread != thread) {
			throw new AffinityException("Processor is assigned to another thread: " + this.thread);
		}
		if (!threadToProcessor.containsKey(thread)) {
			throw new AffinityException("Thread is not assigned to any processor!");
		}
		this.isAssigned = false;
		this.thread = null;
		threadToProcessor.remove(thread);
	}

	public boolean sameChip(Processor other) {
		return other.chipId == this.chipId;
	}

	public boolean sameCore(Processor other) {
		// core ids will be the SAME per chip... so be careful...
		return sameChip(other) && other.coreId == this.coreId;
	}

	public boolean sameChipDifferentCore(Processor other) {
		return sameChip(other) && other.coreId != this.coreId;
	}

	@Override
	public int hashCode() {
		return procId;
	}

	static Processor getAssignedProcessorForThread(Thread t) {
		return threadToProcessor.get(t);
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof Processor) {
			Processor lp = (Processor) o;
			return lp.procId == this.procId;
		}
		return false;
	}

	@Override
	public String toString() {
		return name;
	}

	public int getMask() {
		return 1 << procId;
	}
}
