/*
 * MentaAffinity => http://mentaaffinity.soliveirajr.com
 * Copyright (C) 2012  Sergio Oliveira Jr. (sergio.oliveira.jr@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package org.mentaaffinity.sample;

import org.mentaaffinity.Affinity;

/**
 * If your architecture supports hyper-threading but you don't want another thread sharing the core with your high-priority thread
 * you can assign a thread to the WHOLE core. That means your thread will run alone in the core and the other logical processor will remain
 * idle until the thread exits and calls unbind. Besides not having to compete with any other thread for the whole core, another
 * advantage of doing that is that the thread can have the L1 cache all to itself.
 * 
 * @author Sergio Oliveira Jr.
 */
public class BindToWholeCore {

	public static void main(String[] ags) throws Exception {

		// first you have to mark all threads that you want to control the affinity
		// this must be done inside the thread, in other words, inside the run method

		Thread thread = new Thread(new Runnable() {
			@Override
			public void run() {
				Affinity.bind();
				try {
					while(true) {
						// do whatever you have to do here...
					}
				} finally {
					Affinity.unbind();
				}
			}
		}, "MyHungryThread");

		// after all your threads are marked, then you set the affinity however you want
		// of course you must set the affinity BEFORE the bind method above is called

		Affinity.assignToAnyFreeCore(thread); // you can also pass a chipId to choose the chip...
		Affinity.printSituation();

		// OUTPUT:
		//		Chip-0:
		//		    Core-0.0:
		//		        Processor-0: assigned to MyHungryThread (not-started)
		//		        Processor-1: left idle by MyHungryThread
		//		    Core-1.0:
		//		        Processor-2: free
		//		        Processor-3: free

		thread.start();
		Affinity.printSituation();

		// OUTPUT:
		//		Chip-0:
		//		    Core-0.0:
		//		        Processor-0: bound to MyHungryThread (running)
		//		        Processor-1: left idle by MyHungryThread
		//		    Core-1.0:
		//		        Processor-2: free
		//		        Processor-3: free		
	}
}
